package com.tzld.piaoquan.recommend.server.service;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.Cursor;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.ScanOptions;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.TimeUnit;

/**
 * @author dyp
 */
@Service
@Slf4j
public class PreViewedService {
    @Autowired
    private RedisTemplate<String, Long> redisTemplate;

    public static final String KEY_PRE_VIEWED_FORMAT = "previewed:videos:%s:%s";

    private LoadingCache<String, Set<Long>> preViewedCache = CacheBuilder.newBuilder()
            .maximumSize(100000) // TODO 评估容量
            .refreshAfterWrite(10, TimeUnit.SECONDS)
            .expireAfterWrite(10, TimeUnit.SECONDS)
            .expireAfterAccess(10, TimeUnit.SECONDS)
            .build(new CacheLoader<String, Set<Long>>() {
                @Override
                public Set<Long> load(String key) throws Exception {
                    Set<Long> result = new HashSet<>();
                    Cursor<Long> cursor = redisTemplate.opsForSet().scan(key, ScanOptions.NONE);
                    try {
                        while (cursor != null && cursor.hasNext()) {
                            Long videoId = cursor.next();
                            result.add(videoId);
                        }
                    } catch (Exception e) {
                        log.error("sScan", e);
                    } finally {
                        try {
                            if (!cursor.isClosed()) {
                                cursor.close();
                            }
                        } catch (IOException e) {
                            log.error("sScan", e);
                        }
                    }
                    return result;
                }
            });

    public void updateCache(int appType, String mid, Long[] videoIds) {
        if (StringUtils.isBlank(mid)
                || videoIds == null
                || videoIds.length == 0) {
            return;
        }
        String key = String.format(KEY_PRE_VIEWED_FORMAT, appType, mid);
        redisTemplate.opsForSet().add(key, videoIds);
        redisTemplate.expire(key, 30, TimeUnit.MINUTES);
    }

    public Set<Long> getVideoIds(int appType, String mid) {
        if (StringUtils.isBlank(mid)) {
            return Collections.emptySet();
        }
        String key = String.format(KEY_PRE_VIEWED_FORMAT, appType, mid);
        try {
            return preViewedCache.get(key);
        } catch (Exception e) {
            log.error("get pre viewed cache error! key={}", key, e);
        }
        return Collections.emptySet();
    }

}
