package com.tzld.piaoquan.recommend.server.service.recall;

import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.RegionRelative24HRecallStrategy;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.RegionRelative24HDupRecallStrategy;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.Region24HRecallStrategy;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.RegionHRecallStrategy;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.util.*;
import java.util.concurrent.*;

/**
 * @author dyp
 */
@Service
@Slf4j
public class SpecialRegionRecallStrategy implements ApplicationContextAware, RecallStrategy {

    private final Map<String, RecallStrategy> strategyMap = new HashMap<>();

    @Autowired
    private RedisTemplate<String, String> redisTemplate;

    private ApplicationContext applicationContext;

    @ApolloJsonValue("city_code")
    private Set<String> cityCodes;

    private ExecutorService pool;

    @PostConstruct
    public void init() {
        // init thread pool
        pool = new ThreadPoolExecutor(8, 32,
                0L, TimeUnit.SECONDS, new LinkedBlockingQueue<>(1000),
                new ThreadFactoryBuilder().setNameFormat("video-title-%d").build(), new ThreadPoolExecutor.AbortPolicy());

        Map<String, RecallStrategy> type = applicationContext.getBeansOfType(RecallStrategy.class);
        for (Map.Entry<String, RecallStrategy> entry : type.entrySet()) {
            RecallStrategy value = entry.getValue();
            strategyMap.put(value.getClass().getSimpleName(), value);
        }

    }

    public List<Video> recall(RecallParam param) {

        // 1. region recall
        String provinceCode = StringUtils.isNotBlank(param.getProvinceCode())
                ? param.getProvinceCode()
                : "-1";
        String cityCode = StringUtils.isNotBlank(param.getCityCode())
                ? param.getCityCode()
                : "-1";
        String regionCode;
        if (cityCodes.contains(cityCode)) {
            regionCode = cityCode;
        } else {
            regionCode = provinceCode;
        }

        List<RecallStrategy> strategies = new ArrayList<>();
        if (regionCode.equals("-1")) {
            strategies.add(strategyMap.get(RegionRelative24HRecallStrategy.class.getSimpleName()));
            strategies.add(strategyMap.get(RegionRelative24HDupRecallStrategy.class.getSimpleName()));
        } else {
            strategies.add(strategyMap.get(RegionHRecallStrategy.class.getSimpleName()));
            strategies.add(strategyMap.get(Region24HRecallStrategy.class.getSimpleName()));
            strategies.add(strategyMap.get(RegionRelative24HRecallStrategy.class.getSimpleName()));
            strategies.add(strategyMap.get(RegionRelative24HDupRecallStrategy.class.getSimpleName()));
        }

        // execute
        CountDownLatch cdl = new CountDownLatch(strategies.size());
        List<Future<List<Video>>> recallResultFutures = new ArrayList<>(strategies.size());
        for (RecallStrategy strategy : strategies) {
            Future<List<Video>> future = pool.submit(() -> {
                List<Video> result = strategy.recall(param);
                cdl.countDown();
                return result;
            });
            recallResultFutures.add(future);
        }
        try {
            cdl.await(3000, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            log.error("rov_pool_recall_with_region recall error", e);
            return null;
        }

        // TODO 理解业务后重构，可以先召回 再过滤
        // 2. 去重
        Set<Long> videoIds = new HashSet<>();
        List<Video> results = new ArrayList<>();
        for (Future<List<Video>> f : recallResultFutures) {
            if (results.size() >= param.getSize()) {
                break;
            }
            try {
                List<Video> datas = f.get();
                if (CollectionUtils.isNotEmpty(f.get())) {
                    for (Video data : datas) {
                        if (!videoIds.contains(data.getVideoId())) {
                            videoIds.add(data.getVideoId());
                            results.add(data);
                        }
                        if (results.size() >= param.getSize()) {
                            break;
                        }
                    }
                }
            } catch (Exception e) {
                // log
            }
        }


        // 3. 对在流量池中存在的视频添加标记字段
        for (Video data : results) {

            Long videoId = data.getVideoId();
            String quick_flow_pool_isin_flow_pool_key =
                    String.format("flow:pool:quick:video:ids:%s:3", param.getAppType());
            String quick_flow_pool_flow_pool_key =
                    String.format("flow:pool:quick:video:%s:3:%s", param.getAppType(), videoId);
            if (redisTemplate.opsForSet().isMember(quick_flow_pool_isin_flow_pool_key, videoId)) {
                data.setFlowPool(redisTemplate.opsForSet().randomMember(quick_flow_pool_flow_pool_key));
                data.setInFlowPool(true);
            } else {
                String isIn_flow_pool_key =
                        String.format("flow:pool:video:ids:%s:3", param.getAppType());
                String flow_pool_key =
                        String.format("flow:pool:video:%s:%s", param.getAppType(), videoId);
                if (redisTemplate.opsForSet().isMember(isIn_flow_pool_key, videoId)) {
                    data.setFlowPool(redisTemplate.opsForSet().randomMember(flow_pool_key));
                    data.setInFlowPool(true);
                }
            }
        }

        return results;
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        this.applicationContext = applicationContext;
    }
}
